// Copyright (C) 2009-2010 Ferdinand Majerech
// This file is part of MiniINI
// For conditions of distribution and use, see copyright notice in LICENSE.txt

#ifndef INIFILE_H_INCLUDED
#define INIFILE_H_INCLUDED

#include "typedefs.h"
#include "globals.h"
#include "allocator.h"
#include "inisection.h"
#include "miniini_assert.h"

///INI File
/**
 * Reads and stores data from an ini file and gives access to its sections.
 */
class INIFile
{
    private:

        ///Number of sections in the file. Includes default section so it's at least 1 after INIFile is initialized.
        miniini_private::ui Length;
        ///Array of pointers to sections.
        INISection * * Sections;
        ///Allocator for character data of this file.
        miniini_private::Allocator * Alloc;
        ///Iteration index.
        miniini_private::i Iter;


    public:

        ///Empty constructor.
        INIFile()
            :Length(0)
            ,Sections(NULL)
            ,Alloc(NULL)
            ,Iter(-1)
        {}

        ///Resets iteration
        void Reset()
        {
            Iter = -1;
        }

        ///Advance iteration to the next section.
        /**
         * @return true if there is another section and internal index now points to it.
         * @return false if there are no more sections, i.e. we're finished with iteration.
         */
        bool Next()
        {
            if(Iter < static_cast<miniini_private::i>(Length) - 1)
            {
                ++Iter;
                return true;
            }
            return false;
        }

        ///Loads given INI file.
        /**
		 * @note You can only call OpenFile on an INIFile that is not initialised, i.e. OpenFile or LoadBuffer was not called before.
         * @note File will be loaded even if it contains no ini sections, and even empty sections will be read.
         * @param fname Filename of the file to load. Must be a valid, zero terminated string.
         * @return true if the file is succesfully loaded.
         * @return false if the file does not exist, can not be accessed or is otherwise invalid.
         */
        bool OpenFile(const char * const fname);

        ///Loads given INI file. (STL version)
        /**
		 * @note You can only call LoadBuffer on an INIFile that is not initialised, i.e. OpenFile or LoadBuffer was not called before.
         * @note File will be loaded even if it contains no ini sections, and even empty sections will be read.
         * @param fname Filename of the file to load.
         * @return true if the file is succesfully loaded.
         * @return false if the file does not exist, can not be accessed or is otherwise invalid.
         */
        #ifndef MINIINI_NO_STL
        bool OpenFile (const std::string & fname)
        {
            return OpenFile(fname.c_str());
        }
        #endif

        ///Loads INI file from given zero terminated buffer.
        /**
         * @note File will be loaded even if it contains no ini sections, and even empty sections will be read.
         * Useful e.g. for loading INI files from compressed files.
         * @param buf Buffer to load from. Must be zero terminated.
         * @param size Size of the buffer (including terminating zero).
		 * @return true if successfully loaded.
		 * @return false if there was an error and the buffer was not loaded.
         */
        bool LoadBuffer(const char * buf, unsigned size);

        ///Destructor.
        /**
         * Deletes loaded INI file (if any)
         */
        ~INIFile();

        ///Gets pointer to requested section.
        /**
         * @param name Name of the section to get. Must be a valid, zero terminated string.
         * @return pointer to requested section if the section exists.
         * @return NULL if the section does not exist.
         */
        INISection * GetSection(const char * const name) const;

        ///Gets pointer to requested section. (STL version)
        /**
         * @param name Name of the section to get.
         * @return pointer to requested section if the section exists.
         * @return NULL if the section does not exist.
         */
        #ifndef MINIINI_NO_STL
        INISection * GetSection(const std::string & name) const
        {
            return GetSection(name.c_str());
        }
        #endif

        ///[] operator : alias for GetSection().
        /**
         * @param name Name of the section to get. Must be a valid, zero terminated string.
         * @return pointer to requested section if the section exists.
         * @return NULL if the section does not exist.
         */
        INISection * operator [] (const char * const name) const
        {
            MINIINI_ASSERT(name, "NULL pointer was passed as section name to"
                                 "INIFile::operator []");
            return GetSection(name);
        }

        ///[] operator : alias for GetSection() (STL version).
        /**
         * @param name Name of the section to get.
         * @return pointer to requested section if the section exists.
         * @return NULL if the section does not exist.
         */
        #ifndef MINIINI_NO_STL
        INISection * operator [] (const std::string & name) const
        {
            return GetSection(name);
        }
        #endif

        ///@return number of sections in the file.
        unsigned GetLength() const
        {
            return static_cast<unsigned>(Length);
        }

        ///@return true if this INIFile() is initialised, false otherwise.
        bool IsValid() const
        {
            //If any of these is 0/NULL, this INIFile was constructed
            //but not initialised. (Length will be at least 1 after 
            //initialization due to default section)
            return Length && Sections && Alloc;
        }

    private:

        ///Tries to read a header from line starting by given character pointer.
        /**
         * Updates character pointer to point to next line.
         * If needed, reallocates header buffer and updates its pointer and capacity.
         * @param currentcharref Reference to pointer to current character in a buffer.
         * @param header Reference to buffer where to read the header name (if any) to.
         * @param headercap Reference to the capacity of header buffer.
         * @return size of header or 0 if there is no header in the line or its size is 0 (which is invalid)
         */
        static inline miniini_private::ui Header(const miniini_private::c * & currentcharref, 
                                                 miniini_private::c * & header, 
                                                 miniini_private::ui & headercap);

        INIFile(const INIFile &);

        void operator = (const INIFile &);
};

#endif
